(function(window) {
	var Piece = function(canvas, config)
	{
		this.initialize(canvas, config);
	}
	var p = Piece.prototype = new BasePiece();
	//
	p.initialize = function(canvas, config)
	{
		BasePiece.prototype.initialize.call(this, canvas, config);
		this.initInteraction();
		this.context = this.canvas.getContext('2d');
		this.context.imageSmoothingEnabled = false;
	}

	p.onKeyUp = function(e)
	{
		BasePiece.prototype.onKeyUp.call(this, e);
		if (!this.config.debug) return;
		var c = String.fromCharCode(e.which);
		if (c=="R") this.reset();
		else if (c=="D") this.drawTestLines();
	}
	
	p.drawTestLines = function()
	{
		//Antialiasing issue: http://jsfiddle.net/HDF3a/240/
		//https://stackoverflow.com/questions/7615009/disable-interpolation-when-scaling-a-canvas
		this.drawing = false;
		var w = this.width, h = this.height;
		// plotLine(0,0,411-96,247-80,this.context);
		// plotLine(46,40,411-50,247-40,this.context);
		plotLine(0,0,1200,800,this.context);
	}
	
	
	/*********************************
	 *					INTERACTION
	 ********************************/
	
	p.initInteraction = function()
	{
		if (isTouch)
		{
			this.stage.addEventListener("stagemousedown", this.handleMouseDownTouch.bind(this));
			this.stage.addEventListener("stagemousemove", this.handleMouseMoveTouch.bind(this));
			this.stage.addEventListener("stagemouseup", this.handleMouseUpTouch.bind(this));
		}
		else
		{
			this.stage.addEventListener("stagemousedown", this.handleMouseDown.bind(this));
			this.stage.addEventListener("stagemousemove", this.handleMouseMove.bind(this));
		}
	}
	p.handleMouseDown = function(e)
	{
		var x = e.stageX, y = e.stageY;
		if (!this.drawing)
		{
			this.drawing = true;
		}
		else
		{
			var dsq = Math.sqrt(Math.pow(this.dragStart.x-x,2)+Math.pow(this.dragStart.y-y,2));
			if (dsq>this.config.minimalGradientWidth)
			{
				this.drawGradient(this.dragStart.x,this.dragStart.y,x,y);
			}
		}
		this.dragStart = new Point(x,y);
	}
	p.handleMouseMove = function(e)
	{
		var x = e.stageX, y = e.stageY;
		this.dragPos = new Point(x,y); 
	}
	
	p.handleMouseDownTouch = function(e)
	{
		if (this.drawing) return;
		this.pointerID = e.pointerID;
		var x = e.stageX, y = e.stageY;
		this.drawing = true;
		this.dragStart = this.dragPos = new Point(x,y);
		this.mouseDownAt = Date.now();
		log("down",this.mouseDownAt, this.dragStart.x+","+this.dragStart.y ,this.firstTap?this.firstTap.x+","+this.firstTap.y:"");
	}
	p.handleMouseMoveTouch = function(e)
	{
		if (e.pointerID!=this.pointerID || !this.drawing) return;
		var x = e.stageX, y = e.stageY;
		this.dragPos = new Point(x,y); 
	}
	p.handleMouseUpTouch = function(e)
	{
		if (e.pointerID!=this.pointerID) return;
		var x = e.stageX, y = e.stageY, p = new Point(x,y), cfg = this.config;
		var d = Math.sqrt(Math.pow(this.dragStart.x-x,2)+Math.pow(this.dragStart.y-y,2));
		var dmin = cfg.minimalGradientWidth*this.dpr;
		log("up",d, Date.now()-this.mouseDownAt, this.dragStart.x+","+this.dragStart.y,this.firstTap?this.firstTap.x+","+this.firstTap.y:"");
		this.drawing = false;
		var draw = null;
		if (d>dmin)
		{
			draw = p; 
		}
		else if (cfg.tapEnabled && Date.now()-this.mouseDownAt < 300)
		{
			if (this.firstTap) 
			{
				d = Math.sqrt(Math.pow(this.firstTap.x-x,2)+Math.pow(this.firstTap.y-y,2));
				if (d>dmin) draw = this.firstTap;
			}
			else 
			{
				this.firstTap = this.dragStart;
			}
		}
		if (draw)
		{
			this.drawGradient(this.dragStart.x,this.dragStart.y,draw.x,draw.y);
			this.firstTap = cfg.tapContinuous ? p : null;
		}
	}
	

	 
	/*********************************
	 *					FLOW
	 ********************************/
	
	p.setSize = function(w,h,dpr)
	{
		this.dpr = dpr;
		w = this.width = Math.floor(w*dpr);
		h = this.height = Math.floor(h*dpr);
		//
		this.context.imageSmoothingEnabled = false;
		if (this.tickLast) this.reset();
	}

	p.start = function()
	{
		BasePiece.prototype.start.apply(this);
		log("start",this.width, this.height, this.dpr);
		this.shapeGradient = this.stage.addChild(new Shape());
		this.shapeLine = this.stage.addChild(new Shape());
		this.stage.autoClear = false;
		if (this.width) this.reset();
	}
	
	p.reset = function()
	{
		var cfg = this.config, w = this.width, h = this.height, w2 = w/2, h2 = h/2;
		//draw first, fixed gradient
		var r2 = cfg.initialGradientWidth/2 * this.dpr;
		this.drawGradient(w2-r2,h2,w2+r2,h2);
	}
	
	p.update = function()
	{
		this.stage.clear();
		this.shapeGradient.draw(this.context);
		this.shapeLine.graphics.c();
		if (this.drawing) 
		{
			this.drawLine(this.dragStart.x,this.dragStart.y,this.dragPos.x,this.dragPos.y);
		}
		return false;
	}	
		

	p.drawGradient = function(x0,y0,x1,y1)
	{
		var cfg = this.config;
		var c0 = RandomUtil.pick(cfg.colors);
		var c1 = RandomUtil.pick(cfg.colors, [c0]);
		this.shapeGradient.graphics.c().lf([c0,c1],[0,1],x0,y0,x1,y1).r(0,0,this.width,this.height);
	}
	
	p.drawLine = function(x0,y0,x1,y1)
	{
		if (!this.config.lineAntiAlias)
		{
			this.context.fillStyle = 'black';
			plotLine(x0,y0,x1,y1, this.context);
		}
		else
		{
			this.shapeLine.graphics.c().s("black").ss(1).mt(x0,y0).lt(x1,y1);
			this.shapeLine.draw(this.context);
		}
	}
	
	
	//Adapted from http://members.chello.at/easyfilter/bresenham.js
	//Original will not always finish when coords are not integer!
	var plotLine = function(x0, y0, x1, y1, context)
	{
		var rnd = Math.round;
		x0 = rnd(x0);
		y0 = rnd(y0);
		x1 = rnd(x1);
		y1 = rnd(y1);
		var dx =  Math.abs(x1-x0), sx = x0<x1 ? 1 : -1;
		var dy = -Math.abs(y1-y0), sy = y0<y1 ? 1 : -1;
		var err = dx+dy, e2;
		var z= 1;
		for (;;){
			context.fillRect(x0*z,y0*z,z,z);
			if (x0 == x1 && y0 == y1) break;
			e2 = 2*err;
			if (e2 >= dy) { err += dy; x0 += sx; } 
			if (e2 <= dx) { err += dx; y0 += sy; }
		}
	}

	window.Piece = Piece;
	

}(window));

